from .icon_store import LazyLoadingListStore
from .page_browser import PagePreviewList
from .window import MainWindow
from .dialogs import AboutDialog, OpenDialog

__all__ = ['LazyLoadingListStore', 'PagePreviewList', 'MainWindow', 'AboutDialog', 'OpenDialog']



"""
Fixes warnings about unspecified version contraints a la:
PyGIWarning: GtkSource was imported without specifying a version first. Use gi.require_version('GtkSource', '4') before import to ensure that the right version gets loaded.

import with:

  import tests

in your Gtk using test file in the first line
"""
# pylint: disable=unused-import
from unittest import TestCase
import gi
from pathlib import Path
gi.require_version('Gtk', '3.0')
gi.require_version('GdkPixbuf', '2.0')
try:
    gi.require_version('GtkSource', '4')
except ValueError:
    gi.require_version('GtkSource', '3.0')

ASSETS_PATH = (Path(__file__).parent/ 'assets').absolute()
if not ASSETS_PATH.exists():
    raise RuntimeError('Assset path {} not found, please run: make tests/assets first')





from gi.repository import Gtk, Pango, GObject, Gdk
from enum import Enum
from typing import List, Tuple, Any
from ocrd_utils.constants import MIMETYPE_PAGE, MIME_TO_EXT
from ocrd_browser.model import Document, Page


class Configurator(Gtk.Widget):
    def __init__(self):
        self.document = None
        self.value = None

    def set_document(self, document: Document):
        self.document = document

    def set_value(self, value):
        raise NotImplementedError('You have to override set_value')


class View:
    def __init__(self, name: str, window: Gtk.Window):
        self.name: str = name
        self.window: Gtk.Window = window
        self.document: Document = None
        self.current: Page = None
        self.page_id: str = None

        self.configurators: List[Tuple[str, Configurator]] = []
        self.container: Gtk.Box = None
        self.action_bar: Gtk.ActionBar = None
        self.viewport: Gtk.Viewport = None

    def build(self):
        self.container = Gtk.Box(visible=True, orientation="vertical")
        self.action_bar = Gtk.ActionBar(visible=True)
        self.action_bar.pack_end(CloseButton(self.name))

        scroller = Gtk.ScrolledWindow(visible=True, shadow_type='in', propagate_natural_width=True)
        self.viewport = Gtk.Viewport(visible=True, hscroll_policy='natural', vscroll_policy='natural')
        self.viewport.connect('size-allocate', self.on_viewport_size_allocate)
        scroller.add(self.viewport)

        self.container.pack_start(self.action_bar, False, True, 0)
        self.container.pack_start(scroller, True, True, 0)

    def set_document(self, document: Document):
        self.document = document
        for (_name, configurator) in self.configurators:
            configurator.set_document(document)

    def add_configurator(self, name: str, configurator: Configurator) -> None:
        """
        Adds a configurator for property self.{name}

        @param name: str
        @param configurator: Configurator|Gtk.Widget
        """
        if not hasattr(self, name):
            raise AttributeError(f'{self.__class__.__name__} has no attribute {name}.')
        configurator.set_value(getattr(self, name))
        self.configurators.append((name, configurator))
        configurator.connect('changed', lambda _source, *value: self.config_changed(name, value))
        self.action_bar.pack_start(configurator)

    def config_changed(self, name: str, value: Any) -> None:
        if len(value) == 1:
            value = value[0]
        setattr(self, name, value)

    @property
    def use_file_group(self) -> str:
        return 'OCR-D-IMG'

    def page_activated(self, sender, page_id):
        self.page_id = page_id
        self.reload()

    def pages_selected(self, _sender: Gtk.Widget, page_ids: List[str]) -> None:
        pass

    def reload(self):
        self.current = self.document.page_for_id(self.page_id, self.use_file_group)
        self.redraw()

    def redraw(self):
        pass

    def on_viewport_size_allocate(self, _sender: Gtk.Widget, rect: Gdk.Rectangle):
        self.on_size(rect.width, rect.height, rect.x, rect.y)

    def on_size(self, w, h, x, y):
        pass


class PageQtySelector(Gtk.Box, Configurator):

    def __init__(self):
        super().__init__(visible=True, spacing=3)
        self.value = None

        label = Gtk.Label(label='# Pages:', visible=True)

        self.pages = Gtk.SpinButton(visible=True, max_length=2, width_chars=2, max_width_chars=2, numeric=True)
        self.pages.set_tooltip_text('How many pages should be displayed at once?')
        self.pages.set_adjustment(Gtk.Adjustment.new(1, 1, 16, 1, 4, 3))

        self.pack_start(label, False, True, 0)
        self.pack_start(self.pages, False, True, 0)

        self.pages.connect('value-changed', self.value_changed)

    def set_value(self, value):
        self.value = value
        self.pages.set_value(value)

    def value_changed(self, spin: Gtk.SpinButton):
        self.emit('changed', int(spin.get_value()))

    @GObject.Signal(arg_types=(int,))
    def changed(self, page_qty: int):
        self.value = page_qty


class FileGroupSelector(Gtk.Box, Configurator):

    def __init__(self, filter_=None, show_mime=False, show_ext=True):
        super().__init__(visible=True, spacing=3)
        self.value = None
        label = Gtk.Label(label='Group:', visible=True)

        self.groups = FileGroupComboBox(filter_, show_mime, show_ext)

        self.pack_start(label, False, True, 0)
        self.pack_start(self.groups, False, True, 0)

        self.groups.connect('changed', self.combo_box_changed)

    def set_value(self, value):
        self.value = value
        active_id = None
        for id_, group, mime, _ext in self.groups.get_model():
            if (value[0] is None or value[0] == group) and (value[1] is None or value[1] == mime):
                active_id = id_
                break
        if active_id:
            self.groups.set_active_id(active_id)

    def set_document(self, document: Document):
        self.groups.set_document(document)
        self.set_value(self.value)

    def combo_box_changed(self, combo: Gtk.ComboBox):
        model = combo.get_model()
        if len(model) > 0:
            row = combo.get_model()[combo.get_active()][:]
            self.emit('changed', row[FileGroupComboBox.COLUMN_GROUP], row[FileGroupComboBox.COLUMN_MIME])

    @GObject.Signal()
    def changed(self, file_group: str, mime_type: str):
        self.value = (file_group, mime_type)


class FileGroupComboBox(Gtk.ComboBox):
    COLUMN_ID = 0
    COLUMN_GROUP = 1
    COLUMN_MIME = 2
    COLUMN_EXT = 3

    def __init__(self, filter_=None, show_mime=False, show_ext=True):
        Gtk.ComboBox.__init__(self, visible=True)
        self.set_model(Gtk.ListStore(str, str, str, str))
        self.filter = filter_
        self.set_id_column(self.COLUMN_ID)

        self.add_renderer(self.COLUMN_GROUP, 120 if show_mime else 200)
        if show_mime:
            self.add_renderer(self.COLUMN_MIME, 80)
        if show_ext:
            self.add_renderer(self.COLUMN_EXT)

    def set_document(self, document: Document):
        self.set_model(FileGroupModel.build(document, self.filter))
        # self.set_active(0)

    def add_renderer(self, column, width=None):
        renderer = Gtk.CellRendererText()
        renderer.props.ellipsize = Pango.EllipsizeMode.MIDDLE
        if width:
            renderer.props.width = width
        self.pack_start(renderer, False)
        self.add_attribute(renderer, "text", column)
        return renderer


class FileGroupModel(Gtk.ListStore):
    def __init__(self, document: Document):
        super().__init__(str, str, str, str)
        for group, mime in document.file_groups_and_mimetypes:
            self.append(('{}|{}'.format(group, mime), group, mime, MIME_TO_EXT.get(mime, '.???')))

    @classmethod
    def build(cls, document: Document, filter_: 'FileGroupFilter' = None) -> 'FileGroupModel':
        model = cls(document)
        model = model.filter_new()
        model.set_visible_func(filter_ if filter_ else FileGroupFilter.ALL)
        return model

    @staticmethod
    def image_filter(model, it, _data):
        return model[it][FileGroupComboBox.COLUMN_MIME].startswith('image/')

    @staticmethod
    def page_filter(model, it, _data):
        return model[it][FileGroupComboBox.COLUMN_MIME] == MIMETYPE_PAGE

    @staticmethod
    def all_filter(_model, _it, _data) -> bool:
        return True


class FileGroupFilter(Enum):
    IMAGE = FileGroupModel.image_filter
    PAGE = FileGroupModel.page_filter
    ALL = FileGroupModel.all_filter


class CloseButton(Gtk.Button):
    def __init__(self, view_name):
        Gtk.Button.__init__(self, visible=True)
        self.set_name('close_{}'.format(view_name))
        self.set_detailed_action_name('win.close_view("{}")'.format(view_name))
        self.set_relief(Gtk.ReliefStyle.NONE)
        self.set_always_show_image(True)
        pixbuf = Gtk.IconTheme.get_default().load_icon('window-close', Gtk.IconSize.SMALL_TOOLBAR,
                                                       Gtk.IconLookupFlags.FORCE_SYMBOLIC)
        image = Gtk.Image(visible=True)
        image.set_from_pixbuf(pixbuf)
        self.set_image(image)



import unittest
from unittest.mock import MagicMock
from ocrd_browser.view import ViewXml
from ocrd_browser.ui import MainWindow

class XmlViewTestCase(unittest.TestCase):

    def setUp(self):
        self.vx = ViewXml('unique',MagicMock(spec=MainWindow))

    def test_can_construct(self):
        self.assertIsNotNone(self.vx)


if __name__ == '__main__':
    unittest.main()



from gi.repository import Gio, Gtk, GLib

import pkg_resources
from typing import List
from ocrd_browser.util.gtk import ActionRegistry
from ocrd_browser.ui import MainWindow, AboutDialog, OpenDialog
from ocrd_browser.view import ViewRegistry


class OcrdBrowserApplication(Gtk.Application):
    def __init__(self) -> None:
        Gtk.Application.__init__(self, application_id='org.readmachine.ocrd-browser',
                                 flags=Gio.ApplicationFlags.HANDLES_OPEN)
        self.actions = ActionRegistry(for_widget=self)
        self.view_registry = ViewRegistry.create_from_entry_points()

    def do_startup(self) -> None:
        Gtk.Application.do_startup(self)
        self.actions.create('new')
        self.actions.create('open')
        self.actions.create('about')
        self.actions.create('quit')
        for entry_point in pkg_resources.iter_entry_points('ocrd_browser_ext'):
            (entry_point.load())(self)

    def do_activate(self) -> None:
        win = self.get_active_window()
        if not win:
            win = MainWindow(application=self)
        win.present()

    def on_about(self, _action: Gio.SimpleAction, _param: str = None) -> None:
        about_dialog = AboutDialog(application=self, transient_for=self.get_active_window(), modal=True)
        about_dialog.present()

    def on_quit(self, _action: Gio.SimpleAction, _param: str = None) -> None:
        self.quit()

    def on_open(self, _action: Gio.SimpleAction, _param: str = None) -> None:
        open_dialog = OpenDialog(application=self, transient_for=self.get_active_window(), modal=True)
        response = open_dialog.run()
        if response == Gtk.ResponseType.OK:
            self.open_in_window(open_dialog.get_uri(), window=open_dialog.get_transient_for())
        open_dialog.destroy()

    def on_new(self, _action: Gio.SimpleAction, _param: str = None) -> None:
        win = MainWindow(application=self)
        win.present()

    def do_open(self, files: List[Gio.File], file_count: int, hint: str) -> int:
        for file in files:
            self.open_in_window(file.get_uri(), window=None)
        return 0

    def open_in_window(self, uri: str, window: MainWindow = None) -> None:
        if not window or not window.document.empty:
            window = MainWindow(application=self)
        window.present()
        GLib.timeout_add(10, window.open, uri)



from gi.repository import Gtk, GdkPixbuf, Gio, GObject, GLib

from ocrd_browser.model import Document, DEFAULT_FILE_GROUP
from ocrd_browser.view import ViewRegistry, View, ViewImages
from ocrd_browser.util.gtk import ActionRegistry
from .page_browser import PagePreviewList
from pkg_resources import resource_filename
from typing import List, cast


@Gtk.Template(filename=resource_filename(__name__, '../resources/main-window.ui'))
class MainWindow(Gtk.ApplicationWindow):
    __gtype_name__ = "MainWindow"

    header_bar: Gtk.HeaderBar = Gtk.Template.Child()
    page_list_scroller: Gtk.ScrolledWindow = Gtk.Template.Child()
    panes: Gtk.Paned = Gtk.Template.Child()
    current_page_label: Gtk.Label = Gtk.Template.Child()
    view_container: Gtk.Box = Gtk.Template.Child()
    view_menu_box: Gtk.Box = Gtk.Template.Child()

    def __init__(self, **kwargs):
        Gtk.ApplicationWindow.__init__(self, **kwargs)
        self.set_icon(GdkPixbuf.Pixbuf.new_from_resource("/org/readmachine/ocrd-browser/icons/icon.png"))
        self.views = []
        self.current_page_id = None
        self.document = Document.create(emitter=self.emit)

        vstring = GLib.Variant("s", "")

        self.actions = ActionRegistry(for_widget=self)
        self.actions.create('close')
        self.actions.create('goto_first')
        self.actions.create('go_back')
        self.actions.create('go_forward')
        self.actions.create('goto_last')
        self.actions.create('page_remove')
        self.actions.create('page_properties')
        self.actions.create('close_view', param_type=vstring)
        self.actions.create('create_view', param_type=vstring)

        self.page_list = PagePreviewList(self.document)
        self.page_list_scroller.add(self.page_list)
        self.page_list.connect('page_activated', self.on_page_activated)

        for id_, view in self.view_registry.get_view_options().items():
            menu_item = Gtk.ModelButton(visible=True, centered=False, halign=Gtk.Align.FILL, label=view, hexpand=True)
            menu_item.set_detailed_action_name('win.create_view("{}")'.format(id_))
            self.view_menu_box.pack_start(menu_item, True, True, 0)

        self.add_view(ViewImages)

        self.update_ui()

    def on_page_remove(self, _a: Gio.SimpleAction, _p):
        for page_id in self.page_list.get_selected_ids():
            self.document.delete_page(page_id)

    def on_page_properties(self, _a: Gio.SimpleAction, _p):
        print(self.page_list.get_selected_items())

    @Gtk.Template.Callback()
    def on_recent_menu_item_activated(self, recent_chooser: Gtk.RecentChooserMenu):
        app = self.get_application()
        item: Gtk.RecentInfo = recent_chooser.get_current_item()
        app.open_in_window(item.get_uri(), window=self)

    def open(self, uri):
        self.set_title('Loading')
        self.header_bar.set_title('Loading ...')
        self.header_bar.set_subtitle(uri)
        self.update_ui()
        # Give GTK some break to show the Loading message
        GLib.timeout_add(50, self._open, uri)

    def _open(self, uri):
        self.document = Document.load(uri, emitter=self.emit)
        self.page_list.set_document(self.document)

        title = self.document.workspace.mets.unique_identifier if self.document.workspace.mets.unique_identifier else '<unnamed>'
        self.set_title(title)
        self.header_bar.set_title(title)
        self.header_bar.set_subtitle(self.document.workspace.directory)

        for view in self.views:
            view.set_document(self.document)

        if len(self.document.page_ids):
            self.on_page_activated(None, self.document.page_ids[0])

    @property
    def view_registry(self) -> ViewRegistry:
        return cast(MainWindow, self.get_application()).view_registry

    def add_view(self, view_class):
        name = 'view_{}'.format(len(self.views))
        view: View = view_class(name, self)
        view.build()
        view.set_document(self.document)
        self.views.append(view)
        self.connect('page_activated', view.page_activated)
        self.page_list.connect('pages_selected', view.pages_selected)
        self.view_container.pack_start(view.container, True, True, 3)

    def on_page_activated(self, _, page_id):
        if self.current_page_id != page_id:
            self.current_page_id = page_id
            self.emit('page_activated', page_id)

    @GObject.Signal(arg_types=(str,))
    def page_activated(self, page_id):
        index = self.document.page_ids.index(page_id)
        self.current_page_label.set_text('{}/{}'.format(index + 1, len(self.document.page_ids)))
        self.update_ui()
        pass

    @GObject.Signal(arg_types=(str, object,))
    def document_changed(self, subtype: str, page_ids: List[str]):
        self.page_list.document_changed(subtype, page_ids)

    @GObject.Signal()
    def document_saved(self):
        print('saved')

    def update_ui(self):
        can_go_back = False
        can_go_forward = False
        if self.current_page_id and len(self.document.page_ids) > 0:
            index = self.document.page_ids.index(self.current_page_id)
            last_page = len(self.document.page_ids) - 1
            can_go_back = index > 0
            can_go_forward = index < last_page

        self.actions['goto_first'].set_enabled(can_go_back)
        self.actions['go_back'].set_enabled(can_go_back)
        self.actions['go_forward'].set_enabled(can_go_forward)
        self.actions['goto_last'].set_enabled(can_go_forward)

    def on_close(self, _a: Gio.SimpleAction, _p):
        self.destroy()

    def on_goto_first(self, _a: Gio.SimpleAction, _p):
        self.page_list.goto_index(0)

    def on_go_forward(self, _a: Gio.SimpleAction, _p):
        self.page_list.skip(1)

    def on_go_back(self, _a: Gio.SimpleAction, _p):
        self.page_list.skip(-1)

    def on_goto_last(self, _a: Gio.SimpleAction, _p):
        self.page_list.goto_index(-1)
        return 1

    def on_create_view(self, _a, selected_view_id: GLib.Variant):
        view_class = self.view_registry.get_view(selected_view_id.get_string())
        if view_class:
            self.add_view(view_class)

    def on_close_view(self, _action: Gio.SimpleAction, view_name: GLib.Variant):
        for view in self.views:
            if view.name == view_name.get_string():
                view.container.destroy()
                break
        self.disconnect_by_func(view.page_activated)
        self.views.remove(view)
        del view



from gi.repository import Gtk

from itertools import zip_longest
from ocrd_browser.util.image import pil_to_pixbuf, pil_scale
from .base import View, FileGroupSelector, FileGroupFilter, PageQtySelector


class ViewImages(View):
    """
    View of one or more consecutive images
    """

    label = 'Image'

    def __init__(self, name, window, **kwargs):
        super().__init__(name, window, **kwargs)
        self.file_group = ('OCR-D-IMG', None)
        self.page_qty = 1
        self.preview_height = 10
        self.image_box = None
        self.pages = []

    def build(self):
        super(ViewImages, self).build()

        self.add_configurator('file_group', FileGroupSelector(FileGroupFilter.IMAGE))
        self.add_configurator('page_qty', PageQtySelector())

        self.image_box = Gtk.Box(visible=True, homogeneous=True)
        self.viewport.add(self.image_box)
        self.rebuild_images()

    def config_changed(self, name, value):
        super(ViewImages, self).config_changed(name, value)
        if name == 'page_qty':
            self.rebuild_images()
        self.reload()

    def rebuild_images(self):
        existing_images = {child.get_name(): child for child in self.image_box.get_children()}

        for i in range(0, self.page_qty):
            name = 'image_{}'.format(i)
            if not existing_images.pop(name, None):
                image = Gtk.Image(name=name, visible=True, icon_name='gtk-missing-image', icon_size=Gtk.IconSize.DIALOG)
                self.image_box.add(image)

        for child in existing_images.values():
            child.destroy()

        self.reload()

    def page_activated(self, _sender, page_id):
        self.page_id = page_id
        self.reload()

    @property
    def use_file_group(self):
        return self.file_group[0]

    def reload(self):
        if self.document:
            display_ids = self.document.display_id_range(self.page_id, self.page_qty)
            self.pages = []
            for display_id in display_ids:
                self.pages.append(self.document.page_for_id(display_id, self.use_file_group))
        self.redraw()

    def on_size(self, _w, h, _x, _y):
        if abs(self.preview_height - h) > 4:
            self.preview_height = h
            self.redraw()

    def redraw(self):
        if self.pages:
            image: Gtk.Image
            for image, page in zip_longest(self.image_box.get_children(), self.pages):
                if page:
                    thumbnail = pil_scale(page.image, None, self.preview_height - 10)
                    image.set_from_pixbuf(pil_to_pixbuf(thumbnail))
                else:
                    image.set_from_stock('missing-image', Gtk.IconSize.DIALOG)



from gi.repository import Gtk, GdkPixbuf
from pkg_resources import resource_filename
from ocrd_browser import __version__


@Gtk.Template(filename=resource_filename(__name__, '../resources/about-dialog.ui'))
class AboutDialog(Gtk.AboutDialog):
    __gtype_name__ = "AboutDialog"

    def __init__(self, **kwargs):
        Gtk.AboutDialog.__init__(self, **kwargs)
        self.set_logo(GdkPixbuf.Pixbuf.new_from_resource('/org/readmachine/ocrd-browser/icons/logo.png'))
        self.set_version(__version__)


@Gtk.Template(filename=resource_filename(__name__, '../resources/open-dialog.ui'))
class OpenDialog(Gtk.FileChooserDialog):
    __gtype_name__ = "OpenDialog"

    def __init__(self, **kwargs):
        Gtk.FileChooserDialog.__init__(self, **kwargs)

        filter_text = Gtk.FileFilter()
        filter_text.set_name("METS files")
        filter_text.add_mime_type("text/xml")
        self.add_filter(filter_text)

        filter_any = Gtk.FileFilter()
        filter_any.set_name("Any files")
        filter_any.add_pattern("*")
        self.add_filter(filter_any)



from ocrd import Workspace, Resolver
from ocrd_browser.model import Page
from ocrd_browser.util.image import add_dpi_to_png_buffer
from ocrd_modelfactory import page_from_file
from ocrd_models import OcrdFile
from ocrd_models.ocrd_page_generateds import PcGtsType
from ocrd_models.constants import NAMESPACES as NS
from ocrd_utils import pushd_popd
from ocrd_utils.constants import MIME_TO_EXT
from . import DEFAULT_FILE_GROUP

from typing import Optional, Tuple, List, Set, Union, cast, Callable, Any
from collections import OrderedDict
from pathlib import Path
from urllib.parse import urlparse
from lxml.etree import ElementBase as Element
from numpy import array as ndarray

import cv2

EventCallBack = Optional[Callable[[str, Any], None]]


class Document:

    def __init__(self, workspace: Workspace, mets_url: Union[Path, str] = None, resolver: Resolver = None,
                 emitter: EventCallBack = None):
        self.emitter: EventCallBack = emitter
        self.mets_url = mets_url
        self.workspace: Workspace = workspace
        self.resolver: Resolver = resolver if resolver else Resolver()
        self.empty = True

    @classmethod
    def create(cls, directory: str = None, resolver: Resolver = None, emitter: EventCallBack = None) -> 'Document':
        resolver = resolver if resolver else Resolver()
        workspace = resolver.workspace_from_nothing(directory=directory, mets_basename='mets.xml')
        return cls(workspace, None, resolver=resolver, emitter=emitter)

    @classmethod
    def load(cls, mets_url: Union[Path, str] = None, resolver: Resolver = None,
             emitter: EventCallBack = None) -> 'Document':
        if not mets_url:
            return cls.create(None, resolver=resolver, emitter=emitter)
        mets_url = str(mets_url)
        result = urlparse(mets_url)
        if result.scheme == 'file':
            mets_url = result.path

        resolver = resolver if resolver else Resolver()
        workspace = resolver.workspace_from_url(mets_url, download=True)
        doc = cls(workspace, mets_url, resolver=resolver, emitter=emitter)
        doc.empty = False
        return doc

    @property
    def directory(self) -> Path:
        """
        Get workspace directory as a Path object
        """
        return Path(self.workspace.directory)

    def path(self, other: Union[OcrdFile, Path, str]) -> Path:
        """
        Resolves other relative to current workspace
        """
        if isinstance(other, OcrdFile):
            return self.directory.joinpath(other.local_filename)
        elif isinstance(other, Path):
            return self.directory.joinpath(other)
        elif isinstance(other, str):
            return self.directory.joinpath(other)
        else:
            raise ValueError('Unsupported other of type {}'.format(type(other)))

    @property
    def page_ids(self) -> List[str]:
        """
        List of page_ids in this workspace

        TODO: This is unsorted (or sorted in document order) use @ORDER attribute or keep document order in sync with the correct order

        @return: List[str]
        """
        return cast(List[str], self.workspace.mets.physical_pages)

    @property
    def file_groups(self) -> List[str]:
        """
        List of file_groups in this workspace

        @return: List[str]
        """

        return cast(List[str], self.workspace.mets.file_groups)

    @property
    def mime_types(self) -> Set[str]:
        """
        Set with the distinct mime-types in this workspace

        @return: Set[str]
        """
        return {el.get('MIMETYPE') for el in
                self.workspace.mets._tree.findall('mets:fileSec/mets:fileGrp/mets:file[@MIMETYPE]', NS)}

    @property
    def file_groups_and_mimetypes(self) -> List[Tuple[str, str]]:
        """
        A list with the distinct file_group/mimetype pairs in this workspace

        @return: List[Tuple[str,str]]
        """
        distinct_groups: OrderedDict[Tuple[str, str], None] = OrderedDict()
        for el in self.workspace.mets._tree.findall('mets:fileSec/mets:fileGrp[@USE]/mets:file[@MIMETYPE]', NS):
            distinct_groups[(el.getparent().get('USE'), el.get('MIMETYPE'))] = None

        return list(distinct_groups.keys())

    def get_unused_page_id(self, template_page_id: str = 'PAGE_{page_nr}') -> Tuple[str, int]:
        """
        Finds a page_nr that yields an unused page_id for the workspace and returns page_id, page_nr

        @param template_page_id: str
        @return: Tuple[str, int]
        """
        page_nr = len(self.page_ids) + 1
        page_ids = self.page_ids
        while page_nr < 9999:
            page_id = template_page_id.format(**{'page_nr': page_nr})
            if page_id not in page_ids:
                return page_id, page_nr
            page_nr += 1

        raise RuntimeError('No unused page_id found')

    def display_id_range(self, page_id: str, page_qty: int) -> List[str]:
        """
        Calculates an page_id range of size page_qty around page_id
        @param page_id:
        @param page_qty:
        @return:
        """
        if not page_id:
            return []
        try:
            index = self.page_ids.index(page_id)
        except ValueError:
            return []
        index = index - index % page_qty
        return self.page_ids[index:index + page_qty]

    def page_for_id(self, page_id: str, file_group: str = None) -> Optional['Page']:
        page_file = self.file_for_page_id(page_id, file_group)
        if not page_file:
            return None
        pcgts = self.page_for_file(page_file)
        page = pcgts.get_Page()
        image, info, exif = self.workspace.image_from_page(page, page_id)
        return Page(page_id, page_file, pcgts, image, exif)

    def file_for_page_id(self, page_id: str, file_group: str = DEFAULT_FILE_GROUP) -> Optional[OcrdFile]:
        with pushd_popd(self.workspace.directory):
            files = self.workspace.mets.find_files(fileGrp=file_group, pageId=page_id)
            if not files:
                return None
            return self.workspace.download_file(files[0])

    def page_for_file(self, page_file: OcrdFile) -> PcGtsType:
        with pushd_popd(self.workspace.directory):
            return page_from_file(page_file)

    def reorder(self, ordered_page_ids: List[str]) -> None:
        """
        Orders the pages in physSequence according to ordered_page_ids

        """
        if set(self.page_ids) != set(ordered_page_ids):
            raise ValueError('page_ids do not match: missing in mets.xml: {} / missing in order: {}'.format(
                set(ordered_page_ids).difference(set(self.page_ids)),
                set(self.page_ids).difference(set(ordered_page_ids))
            ))

        page_sequence: Element = self.workspace.mets._tree.getroot().xpath(
            '/mets:mets/mets:structMap[@TYPE="PHYSICAL"]/mets:div[@TYPE="physSequence"]',
            namespaces=NS)[0]

        ordered_divs = []
        for page_id in ordered_page_ids:
            divs = page_sequence.xpath('mets:div[@TYPE="page"][@ID="%s"]' % page_id, namespaces=NS)
            if divs:
                ordered_divs.append(divs[0])
                page_sequence.remove(divs[0])

        if len(page_sequence) > 0:
            raise RuntimeError('page_sequence not empty, still has: {}'.format(page_sequence.getchildren()))

        for div in ordered_divs:
            page_sequence.append(div)

        self._emit('document_changed', 'reordered', self.page_ids)


    def save(self) -> None:
        self.workspace.save_mets()
        self._emit('document_saved')

    def delete_image(self, page_id: str, file_group: str = 'OCR-D-IMG') -> OcrdFile:
        image_files = self.workspace.mets.find_files(pageId=page_id, fileGrp=file_group, local_only=True,
                                                     mimetype='//image/.+')
        # TODO rename to delete_images and cope with it, e.g. for file groups with segementation images
        if len(image_files) != 1:
            print('oh oh', image_files)
            return
        image_file = image_files[0]
        self.workspace.remove_file(image_file, force=False, keep_file=False, page_recursive=True, page_same_group=True)
        self._emit('document_changed', 'page_changed', [page_id])
        return image_file

    def delete_page(self, page_id: str) -> None:
        files = self.workspace.mets.find_files(pageId=page_id, local_only=True)
        for file in files:
            self.workspace.remove_file(file, force=False, keep_file=False)
        self._emit('document_changed', 'page_deleted', [page_id])

    def add_image(self, image: ndarray, page_id: str, file_id: str, file_group: str = 'OCR-D-IMG', dpi: int = 300,
                  mimetype: str = 'image/png') -> 'OcrdFile':
        extension = MIME_TO_EXT[mimetype]
        retval, image_array = cv2.imencode(extension, image)
        image_bytes = add_dpi_to_png_buffer(image_array.tostring(), dpi)
        local_filename = Path(file_group, '%s%s' % (file_id, extension))
        url = (Path(self.workspace.directory) / local_filename)
        current_file = self.workspace.add_file(file_group, ID=file_id, mimetype=mimetype, force=True,
                                               content=image_bytes, url=str(url),
                                               local_filename=str(local_filename), pageId=page_id)
        self.empty = False
        self._emit('document_changed', 'page_added', [page_id])
        return current_file

    def _emit(self, event: str, *args: Any) -> None:
        if self.emitter is not None:
            self.emitter(event,  *args)



from tests import TestCase, ASSETS_PATH
from ocrd_browser.model import Document


class DocumentTestCase(TestCase):

    def setUp(self):
        self.doc = Document.load(ASSETS_PATH / 'kant_aufklaerung_1784/data/mets.xml')

    def test_get_page_ids(self):
        self.assertEqual(['PHYS_0017','PHYS_0020'], self.doc.page_ids)

    def test_reorder(self):
        self.doc.reorder(['PHYS_0020','PHYS_0017'])
        self.assertEqual(['PHYS_0020','PHYS_0017'], self.doc.page_ids)

    def test_reorder_with_wrong_ids_raises_value_error(self):
        with self.assertRaises(ValueError) as context:
            self.doc.reorder(['PHYS_0021','PHYS_0017'])

        self.assertIn('page_ids do not match', str(context.exception))



from ocrd_models import OcrdFile, OcrdExif
from ocrd_models.ocrd_page_generateds import PcGtsType, PageType, MetadataType
from PIL.Image import Image


class Page:
    def __init__(self, id_: str, file: OcrdFile, pc_gts: PcGtsType, image: Image, exif: OcrdExif):
        self.id: str = id_
        self.file: OcrdFile = file
        self.pc_gts: PcGtsType = pc_gts
        self.image: Image = image
        self.exif: OcrdExif = exif

    @property
    def page(self) -> PageType:
        return self.pc_gts.get_Page()

    @property
    def meta(self) -> MetadataType:
        return self.pc_gts.get_Metadata()



import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GdkPixbuf, GObject, GLib

import unittest
from ocrd_browser.ui import LazyLoadingListStore


class LazyLoadingListStoreTestCase(unittest.TestCase):

    def setUp(self):
        def call_instantly(fun, *args, **kwargs):
            return fun(*args, **kwargs)

        self.types = (str, int, GdkPixbuf.Pixbuf)
        self.list_store = Gtk.ListStore(*self.types)
        self.lazy_list_store = LazyLoadingListStore(*self.types, init_row=self.logging_init, load_row=self.logging_load,
                                                    hash_row=self.hash_row)
        self.init_log = []
        self.backup_idle_add = GLib.idle_add
        GLib.idle_add = call_instantly

    def tearDown(self) -> None:
        GLib.idle_add = self.backup_idle_add

    def logging_init(self, row: Gtk.TreeModelRow):
        self.init_log.append(row[:])

    def logging_load(self, row: Gtk.TreeModelRow):
        pass
        # self.init_log.append(row[:])

    def hash_row(self, row: Gtk.TreeModelRow):
        return row[0]

    def test_constructor_propagates_types(self):
        for i in range(0, self.list_store.get_n_columns()):
            self.assertEqual(self.list_store.get_column_type(i), self.lazy_list_store.get_column_type(i))

    def test_constructor_adds_hash_column(self):
        list_width = self.list_store.get_n_columns()
        lazy_list_width = self.lazy_list_store.get_n_columns()
        self.assertEqual(list_width + 1, lazy_list_width)
        hash_column_type = self.lazy_list_store.get_column_type(lazy_list_width - 1)
        str_as_gtype = GObject.GType.from_name('gchararray')
        self.assertEqual(str_as_gtype, hash_column_type)

    def test_append_adds_hash_column(self):
        row_data = ('a', 5, None)
        self.lazy_list_store.append(row_data)
        actual = self.lazy_list_store[-1][:]
        self.assertSequenceEqual(row_data + (None,), actual)

    def test_append_calls_init_row(self):
        row_data = ('a', 5, None)
        self.lazy_list_store.append(row_data)
        self.assertEqual(1, len(self.init_log))
        self.assertSequenceEqual(row_data, self.init_log[0][:len(row_data)])

    def test_init_row_can_modify_model(self):
        def change_row(row: Gtk.TreeModelRow):
            row[1] = 15

        self.lazy_list_store.init_row = change_row
        self.lazy_list_store.append(('a', 5, None))
        self.assertEqual(self.lazy_list_store[-1][1], 15)


if __name__ == '__main__':
    unittest.main()



from typing import Any, Optional, Tuple

from gi.repository import GObject, GtkSource

from ocrd_utils.constants import MIMETYPE_PAGE
from ocrd_models.ocrd_page import to_xml
from ocrd_browser.view import View
from ocrd_browser.view.base import FileGroupSelector, FileGroupFilter

GObject.type_register(GtkSource.View)


class ViewXml(View):
    """
    A view of the current PAGE-XML with syntax highlighting
    """

    label = 'PAGE-XML'

    def __init__(self, name, window, **kwargs):
        super().__init__(name, window, **kwargs)
        self.file_group: Tuple[Optional[str], Optional[str]] = (None, MIMETYPE_PAGE)
        self.text_view: GtkSource.View = None
        self.buffer: GtkSource.Buffer = None

    def build(self):
        super().build()
        self.add_configurator('file_group', FileGroupSelector(FileGroupFilter.PAGE))

        lang_manager = GtkSource.LanguageManager()
        style_manager = GtkSource.StyleSchemeManager()

        self.text_view = GtkSource.View(visible=True, vexpand=False, editable=False, monospace=True,
                                        show_line_numbers=True,
                                        width_request=400)
        self.buffer: GtkSource.Buffer = self.text_view.get_buffer()
        self.buffer.set_language(lang_manager.get_language('xml'))
        self.buffer.set_style_scheme(style_manager.get_scheme('tango'))

        self.viewport.add(self.text_view)

    @property
    def use_file_group(self):
        return self.file_group[0]

    def config_changed(self, name, value):
        super().config_changed(name, value)
        self.reload()

    def redraw(self) -> None:
        if self.current:
            text = to_xml(self.current.pc_gts)
            # TODO: Crashes with big XML, as a workaround shorten the file
            if len(text) > 50000:
                self.buffer.set_highlight_syntax(False)
                line_break_start = text.find("\n", 45000)
                line_break_end = text.find("\n", len(text) - 5000)

                text = text[:line_break_start] + "\n\n\n" + \
                       "... I'm sorry Dave, I'm afraid I can't do that ...\n" + \
                       "... With bigger XML files there are frequent crashes ...\n\n" + \
                       text[line_break_end:]

            else:
                self.buffer.set_highlight_syntax(True)

            self.buffer.set_text(text)



from pkg_resources import EntryPoint, iter_entry_points
from typing import Dict, Tuple, Optional
from .base import View


class ViewRegistry:
    def __init__(self, views):
        self.views: Dict[str, Tuple[str, str, str]] = views

    @classmethod
    def create_from_entry_points(cls) -> 'ViewRegistry':
        views = {}
        entry_point: EntryPoint
        for entry_point in iter_entry_points('ocrd_browser_view'):
            view_class = entry_point.load()
            assert issubclass(view_class, View)
            label = view_class.label if hasattr(view_class, 'label') else view_class.__name__
            description = view_class.__doc__.strip()
            views[entry_point.name] = (view_class, label, description)
        return cls(views)

    def get_view_options(self) -> Dict[str, str]:
        return {id_: label for id_, (view_class, label, description) in self.views.items()}

    def get_view(self, id_) -> Optional[str]:
        return self.views[id_][0] if id_ in self.views else None



from pathlib import Path

from gi.repository import Gtk, Gdk, GdkPixbuf, Pango, GLib, Gio, GObject
from pkg_resources import resource_filename
from typing import List, Optional, Dict
from itertools import count
from ocrd_browser.util.image import cv_scale, cv_to_pixbuf
from ocrd_browser.model import Document, DEFAULT_FILE_GROUP
from . import LazyLoadingListStore

import cv2
import os


@Gtk.Template(filename=resource_filename(__name__, '../resources/page-list.ui'))
class PagePreviewList(Gtk.IconView):
    __gtype_name__ = "PagePreviewList"

    def __init__(self, document: Document, **kwargs):
        super().__init__(**kwargs)
        self.document: Document = None
        self.current: Gtk.TreeIter = None
        self.model: LazyLoadingListStore = None
        self.loading_image_pixbuf: GdkPixbuf.Pixbuf = None
        self.context_menu: Gtk.Menu = None
        self.setup_ui()
        self.setup_context_menu()
        self.set_document(document)

    def setup_context_menu(self):
        menu = Gio.Menu()
        action_menu = Gio.Menu()
        action_menu.append('Remove', 'win.page_remove')
        menu.append_section(None, action_menu)
        prop_menu = Gio.Menu()
        prop_menu.append('Properties', 'win.page_properties')
        menu.append_section(None, prop_menu)
        self.context_menu: Gtk.Menu = Gtk.Menu()
        self.context_menu.bind_model(menu, None, True)
        self.context_menu.attach_to_widget(self)
        self.context_menu.show_all()

    def set_document(self, document):
        self.document = document
        self.setup_model()

    def document_changed(self, subtype: str, page_ids: List[str]):
        def _document_page_added(page_ids: List[str]):
            page_id = None
            for page_id in page_ids:
                file = self.document.workspace.mets.find_files(fileGrp=DEFAULT_FILE_GROUP, pageId=page_id)[0]
                file_name = str(self.document.path(file.local_filename))
                self.model.append((page_id, '', file_name, None, len(self.model)))
            if page_id is not None:
                self.scroll_to_id(page_id)

        def _document_page_deleted(page_ids: List[str]):
            for delete_page_id in reversed(page_ids):
                n, row = self.model.get_row_by_column_value(0, delete_page_id)
                self.model.remove(self.model.get_iter(Gtk.TreePath(n)))

        def _document_page_changed(page_ids: List[str]):
            for page_id in page_ids:
                n, row = self.model.get_row_by_column_value(0, page_id)
                files = self.document.workspace.mets.find_files(fileGrp=DEFAULT_FILE_GROUP, pageId=page_id)
                if files:
                    file_name = str(self.document.path(files[0]))
                    row[2] = file_name

        def _document_reordered(_page_ids: List[str]):
            order = count(start=1)
            for page_id in self.document.page_ids:
                n, row = self.model.get_row_by_column_value(0, page_id)
                row[4] = next(order)

        handler = {
            'page_added': _document_page_added,
            'page_deleted': _document_page_deleted,
            'page_changed': _document_page_changed,
            'reordered': _document_reordered,
        }
        handler[subtype](page_ids)


    def setup_ui(self):
        self.loading_image_pixbuf = GdkPixbuf.Pixbuf.new_from_resource(
            '/org/readmachine/ocrd-browser/icons/loading.png')
        self.set_text_column(0)
        self.set_tooltip_column(1)
        self.set_pixbuf_column(3)
        text_renderer: Gtk.CellRendererText = \
            [cell for cell in self.get_cells() if isinstance(cell, Gtk.CellRendererText)][0]
        text_renderer.props.ellipsize = Pango.EllipsizeMode.MIDDLE
        self.connect('button-release-event', self.button_pressed)

    def button_pressed(self, _sender, event: Gdk.EventButton):
        if event.get_button()[1] == 3:
            path, renderer = self.get_item_at_pos(*event.get_coords())
            row = self.model[path]
            self.emit('on_context_menu', event, row, path, renderer)

    @GObject.Signal(arg_types=(object, object, object, object))
    def on_context_menu(self, event: Gdk.EventButton, row: Gtk.TreeModelRow, path: Gtk.TreePath,
                        renderer: Gtk.CellRenderer):
        if len(self.get_selected_items()) <= 1:
            self.set_cursor(path, None, False)
            self.emit('select_cursor_item')
        self.context_menu.popup_at_pointer(event)

    def setup_model(self):
        file_lookup = self.get_image_paths(self.document)
        self.model = LazyLoadingListStore(str, str, str, GdkPixbuf.Pixbuf, int,
                                          init_row=self.init_row, load_row=self.load_row, hash_row=self.hash_row)
        order = count(start=1)
        for page_id in self.document.page_ids:
            file = str(file_lookup[page_id])
            self.model.append((page_id, '', file, None, next(order)))
        self.set_model(self.model)
        self.model.set_sort_column_id(4, Gtk.SortType.ASCENDING)
        GLib.timeout_add(10, self.model.start_loading)

    def init_row(self, row: Gtk.TreeModelRow):
        row[1] = 'Loading {}'.format(row[2])
        row[3] = self.loading_image_pixbuf

    @staticmethod
    def load_row(row):
        image = cv2.imread(row[2])
        row[1] = '{} ({}x{})'.format(row[2], image.shape[1], image.shape[0])
        row[3] = cv_to_pixbuf(cv_scale(image, 100, None))
        return row

    @staticmethod
    def hash_row(row: Gtk.TreeModelRow):
        file = row[2]
        mtime = os.path.getmtime(file)
        return '{}:{}'.format(file, mtime)

    @staticmethod
    def get_image_paths(document: Document, file_group='OCR-D-IMG') -> Dict[str, Path]:
        images = document.workspace.mets.find_files(fileGrp=file_group)
        page_ids = document.workspace.mets.get_physical_pages(for_fileIds=[image.ID for image in images])
        file_paths = [document.path(image.url) for image in images]
        return dict(zip(page_ids, file_paths))

    def get_selected_ids(self):
        return [self.model[path][0] for path in self.get_selected_items()]

    def goto_index(self, index):
        index = index if index >= 0 else len(self.model) + index
        if 0 <= index < len(self.model):
            self.goto_path(Gtk.TreePath(index))

    def path_for_id(self, page_id):
        n, row = self.model.get_row_by_column_value(0, page_id)
        return Gtk.TreePath(n) if n is not None else None

    def scroll_to_id(self, page_id):
        path = self.path_for_id(page_id)
        self.scroll_to_path(path, False, 0, 1.0)

    def skip(self, pos):
        if not self.current:
            self.current = self.model.get_iter(Gtk.TreePath(0))
        iterate = self.model.iter_previous if pos < 0 else self.model.iter_next
        n = self.current
        for _ in range(0, abs(pos)):
            n = iterate(n)
        self.goto_path(self.model.get_path(n))

    def goto_path(self, path):
        self.set_cursor(path, None, False)
        self.emit('select_cursor_item')
        self.emit('activate_cursor_item')
        self.grab_focus()

    def do_selection_changed(self):
        self.emit('pages_selected', self.get_selected_ids())

    def do_item_activated(self, path: Gtk.TreePath):
        self.current = self.model.get_iter(path)
        self.emit('page_activated', self.model[path][0])

    @GObject.Signal()
    def page_activated(self, page_id: str):
        pass

    @GObject.Signal(arg_types=(object,))
    def pages_selected(self, page_ids: List[str]):
        print(page_ids)
        pass



import io
import cv2
import struct
import zlib

from typing import Tuple, Union
from PIL.Image import Image
from numpy import array as ndarray
from gi.repository import GdkPixbuf

__all__ = ['cv_scale', 'cv_to_pixbuf', 'pil_to_pixbuf', 'pil_scale', 'add_dpi_to_png_buffer']


def cv_to_pixbuf(cv_image: ndarray) -> GdkPixbuf:
    ret, byte_array = cv2.imencode('.jpg', cv_image)
    return bytes_to_pixbuf(byte_array.tobytes())


def pil_to_pixbuf(im: Image) -> GdkPixbuf:
    bytes_io = io.BytesIO()
    im.save(bytes_io, "PNG" if im.mode in ("LA", "RGBA") else "JPEG")
    return bytes_to_pixbuf(bytes_io.getvalue())


def bytes_to_pixbuf(bytes_: bytes) -> GdkPixbuf:
    loader = GdkPixbuf.PixbufLoader()
    loader.write(bytes_)
    loader.close()
    return loader.get_pixbuf()


def cv_scale(orig: ndarray, w: int = None, h: int = None) -> ndarray:
    """
    Scale a cv2 image
    :param orig: ndarray Original cv2 image
    :param w: New width
    :param h: New height
    :return: ndarray
    """
    height, width, depth = orig.shape
    new_width, new_height = _calculate_scale(width, height, w, h)
    return cv2.resize(orig, (new_width, new_height))


def pil_scale(orig: Image, w: int = None, h: int = None) -> ndarray:
    """
    Scale a Pillow image
    :param orig: ndarray Original cv2 image
    :param w: New width
    :param h: New height
    :return: ndarray
    """
    new_width, new_height = _calculate_scale(orig.width, orig.height, w, h)
    thumb = orig.copy()
    thumb.thumbnail((new_width, new_height))
    return thumb


def add_dpi_to_png_buffer(image_bytes: bytes, dpi: Union[int, Tuple[int, int]] = 300) -> bytes:
    """
    adds dpi information to a png image

    see https://stackoverflow.com/questions/57553641/how-to-save-dpi-info-in-py-opencv/57555123#57555123

    """
    if isinstance(dpi, (int, float)):
        dpi = (dpi, dpi)

    # Find start of IDAT chunk
    idat_offset = image_bytes.find(b'IDAT') - 4

    # Create our lovely new pHYs chunk - https://www.w3.org/TR/2003/REC-PNG-20031110/#11pHYs
    phys_chunk = b'pHYs' + struct.pack('!IIc', int(dpi[0] / 0.0254), int(dpi[1] / 0.0254), b"\x01")
    phys_chunk = struct.pack('!I', 9) + phys_chunk + struct.pack('!I', zlib.crc32(phys_chunk))

    return image_bytes[0:idat_offset] + phys_chunk + image_bytes[idat_offset:]


def _calculate_scale(old_width: int, old_height: int, new_width: int = None, new_height: int = None) -> Tuple[int, int]:
    """
    Calculate scaled image size, while keeping the aspect ratio
    :param old_height:int
    :param old_width:int
    :param new_width:int New width
    :param new_height:int New height
    :return: ndarray
    """
    if new_width and new_height:
        raise RuntimeError('Cant scale both width and height')
    if new_height:
        # noinspection PyUnresolvedReferences
        image_scale = new_height / old_height
    elif new_width:
        # noinspection PyUnresolvedReferences
        image_scale = new_width / old_width
    else:
        image_scale = 1

    return int(old_width * image_scale), int(old_height * image_scale)
